package gov.va.rams.service;

import gov.va.rams.entities.Study;
import gov.va.rams.entities.StudyStatus;
import gov.va.rams.entities.Study_Members;
import gov.va.rams.entities.User;
import java.io.File;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.Authenticator;
import java.net.PasswordAuthentication;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.WeakHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.ejb.Singleton;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.codehaus.jettison.json.JSONException;
import org.codehaus.jettison.json.JSONObject;
import org.jasypt.encryption.pbe.StandardPBEStringEncryptor;
import org.codehaus.jettison.json.JSONArray;

import org.apache.commons.codec.binary.Base64;
//import java.util.Base64; //This will be used later in Java 1.8

/**
 * Mediate all SharePoint access.
 * @author Brian
 */
@Singleton
public class SharePoint {
    // There is one base site with many lists and committee sub-sites.
    private String userName, encodedPassword, baseSiteFullUrl;
    private char[] password;
    private String applicationListName, applicationListFacilityLookupColumnName, applicationListAdministratorColumnName;
    private String[] applicationListReviewerColumnNames;
    private String facilityListName, facilityListLiaisonColumnName, facilityListCommitteeLookupColumnName;
    private String committeeListName;
    private String[] committeeListGroupColumnNames;
    // The document site may be another (possibly rolling-yearly) site collection.
    private String documentSiteFullUrlPattern, documentLibraryName, reviewFolderName, studyTeamFolderName;
    // JNDI-based configuration (defaults from META-INF/context.xml)
    @PostConstruct
    private void configure() {
        try {
            Context env = (Context) InitialContext.doLookup("java:comp/env/rams/sp");

            userName = (String) env.lookup("userName");
            encodedPassword = (String) env.lookup("encodedPassword");
    		
    		password = decryptPassword(userName, encodedPassword);
    		
            baseSiteFullUrl = (String) env.lookup("baseSiteUrl");
            applicationListName = (String) env.lookup("applicationListName");
            applicationListReviewerColumnNames = ((String) env.lookup("applicationListReviewerColumnNames")).split(" ");
            try {
                applicationListAdministratorColumnName = (String) env.lookup("applicationListAdministratorColumnName");
            } catch (NamingException ne) {
                Logger.getLogger(SharePoint.class.getName()).log(Level.WARNING, "Missing configuration for SharePoint module set to default", ne);
                applicationListAdministratorColumnName = "AssignedIRBAdministratorId";
            }
            applicationListFacilityLookupColumnName = ((String) env.lookup("applicationListFacilityLookupColumnName"));
            facilityListName = ((String) env.lookup("facilityListName"));
            facilityListLiaisonColumnName = ((String) env.lookup("facilityListLiaisonColumnName"));
            facilityListCommitteeLookupColumnName = ((String) env.lookup("facilityListCommitteeLookupColumnName"));
            committeeListName = ((String) env.lookup("committeeListName"));
            committeeListGroupColumnNames = ((String) env.lookup("committeeListGroupColumnNames")).split(" ");
            documentSiteFullUrlPattern = (String) env.lookup("documentSiteUrlPattern");
            documentLibraryName = (String) env.lookup("documentLibraryName");
            reviewFolderName = (String) env.lookup("reviewFolderName");
            studyTeamFolderName = (String) env.lookup("generatedFolderName");
        } catch (NamingException ne) {
            Logger.getLogger(SharePoint.class.getName()).log(Level.SEVERE, "Missing configuration for SharePoint module", ne);
        }
        boolean reviewersIncludeAdmin = false;
        for (String reviewer : applicationListReviewerColumnNames) {
            reviewersIncludeAdmin = reviewer.equals(applicationListAdministratorColumnName);
            if (reviewersIncludeAdmin)
                break;
        }
        if (!reviewersIncludeAdmin) {
            applicationListReviewerColumnNames = Arrays.copyOf(applicationListReviewerColumnNames, applicationListReviewerColumnNames.length + 1);
            applicationListReviewerColumnNames[applicationListReviewerColumnNames.length - 1] = applicationListAdministratorColumnName;
        }
        Authenticator.setDefault(new Authenticator() {
            @Override
            protected PasswordAuthentication getPasswordAuthentication() {
            	return new PasswordAuthentication(userName, password);
            }
        });
    }

    @PreDestroy
    private void teardown() {
        for (int p = 0; p < password.length; p++) {
			password[p]=0;
		}
        Authenticator.setDefault(null);
    }

    private char[] decryptPassword(String userName, String encodedPassword) {
    	StandardPBEStringEncryptor encryptor = new StandardPBEStringEncryptor();
		encryptor.setPassword(userName); // we HAVE TO set a password
		encryptor.setAlgorithm("PBEWithMD5AndTripleDES"); // optionally set the algorithm

		char[] pwd = encryptor.decrypt(encodedPassword).toCharArray();

		return pwd;
    }

    private String siteApiUrl;
    private void openSession(String siteUrl) throws IOException, JSONException {
        siteApiUrl = siteUrl + "/_api/";
        xRequestDigest = null; // demand-loaded
    }
    private String xRequestDigest;
    private void closeSession() {
        xRequestDigest = null;
        siteApiUrl = null;
    }

    private static final String JSON_ODATA = "application/json;odata=verbose";
    private URLConnection connect(String path) throws IOException {
        URLConnection conn = new URL(siteApiUrl + path.replace(" ", "%20")).openConnection(); // The only special character we have to URI escape is the space.
        conn.setRequestProperty("Accept", JSON_ODATA);
        return conn;
    }

    private void post(URLConnection conn, JSONObject jsonBody, File fileBody, boolean inContextInfo) throws IOException, JSONException {
        assert (jsonBody == null || fileBody == null);
        if (!inContextInfo && xRequestDigest == null) {
            URLConnection ci = connect("ContextInfo");
            post(ci, null, null, true);
            JSONObject ciBody = getBody(ci);
            xRequestDigest = ciBody.getJSONObject("GetContextWebInformation").getString("FormDigestValue");
        }
        if (!inContextInfo)
            conn.setRequestProperty("X-RequestDigest", xRequestDigest);

        if (fileBody == null)
            conn.setRequestProperty("Content-Type", JSON_ODATA);
        else
            conn.setRequestProperty("Content-Length", String.valueOf(Files.size(fileBody.toPath())));
        conn.setDoOutput(true);
        try (OutputStream os = conn.getOutputStream()) {
            if (jsonBody != null) {
                try (OutputStreamWriter osw = new OutputStreamWriter(os, "UTF-8")) {
                    jsonBody.write(osw);
                }
            } else if (fileBody != null) {
                Files.copy(fileBody.toPath(), os);
            }
        }
    }

    private void post(URLConnection conn, JSONObject body) throws IOException, JSONException {
        post(conn, body, null, false);
    }

    private void post(URLConnection conn, File body) throws IOException, JSONException {
        post(conn, null, body, false);
    }

    private void post(URLConnection conn) throws IOException, JSONException {
        post(conn, null, null, false);
    }

    // Get the OData JSON body, unwrapping the outer-level {"d":...} object.
    private static JSONObject getBody(URLConnection conn) throws IOException, JSONException {
        StringBuilder sb = new StringBuilder();
        try (InputStream in = conn.getInputStream()) {
            try (InputStreamReader isr = new InputStreamReader(in, "UTF-8")) {
                for (int c; (c = isr.read()) != -1;)
                    sb.append((char) c);
            }
        }
        return new JSONObject(sb.toString()).getJSONObject("d");
    }

    private static void ignoreBody(URLConnection conn) throws IOException {
        try (InputStream in = conn.getInputStream()) {
            while (true) {
                if (in.read() == -1)
                    break;
            }
        }
    }

    private static final String DOMAIN_USER_PREFIX = "i:0#.w|"; // Windows claims prefix
    private User web_ensureUser(String loginName) throws IOException, JSONException {
        URLConnection conn = connect("Web/EnsureUser");
        JSONObject o = new JSONObject();
        o.put("logonName", DOMAIN_USER_PREFIX + loginName);
        post(conn, o);
        o = getBody(conn);
        User u = new User();
        u.setDisplayName(o.getString("Title"));
        u.setLoginName(o.getString("LoginName").substring(DOMAIN_USER_PREFIX.length()).toLowerCase());
        u.setPermanentLoginIdentifier(o.getJSONObject("UserId").getString("NameId"));
        u.setSPUser(o.getInt("Id"));
        u.setSynchronizedTimestamp(new Date(conn.getDate()));
        return u;
    }

    private String web_getUserById_loginName(int id) throws IOException, JSONException {
        URLConnection conn = connect("Web/GetUserById(" + id + ")/LoginName");
        String name = getBody(conn).getString("LoginName");
        if (name.startsWith(DOMAIN_USER_PREFIX)) // should always be present
            name = name.substring(DOMAIN_USER_PREFIX.length());
        return name;
    }

    public User getUserByLoginName(final String loginName) {
        try {
            openSession(baseSiteFullUrl);
            return web_ensureUser(loginName);
        } catch (IOException | JSONException ex) {
            Logger.getLogger(SharePoint.class.getName()).log(Level.SEVERE, null, ex);
            throw ServiceException.fatal(ex);
        } finally {
            closeSession();
        }
    }

    private void web_folders_add(String serverRelativeUrl) throws IOException, JSONException {
        URLConnection conn = connect("Web/Folders/Add('" + serverRelativeUrl + "')");
        post(conn);
        ignoreBody(conn);
    }

    private void web_folders_files_add(String folderServerRelativeUrl, File file, String asFileName) throws IOException, JSONException {
        URLConnection conn = connect("Web/GetFolderByServerRelativeUrl('" + folderServerRelativeUrl + "')/Files/Add(url='" + asFileName + "',overwrite=true)");
        post(conn, file);
        ignoreBody(conn);
    }

    private static String getFolderPermissionsUrl(String folderServerRelativeUrl) {
        return "Web/GetFolderByServerRelativeUrl('" + folderServerRelativeUrl + "')/ListItemAllFields";
    }

    private void permissions_breakRoleInheritance(String permissionsUrl) throws IOException, JSONException {
        URLConnection conn = connect(permissionsUrl + "/BreakRoleInheritance(copyroleassignments=false)");
        post(conn);
        ignoreBody(conn);
    }

    private void permissions_resetRoleInheritance(String permissionsUrl) throws IOException, JSONException {
        URLConnection conn = connect(permissionsUrl + "/ResetRoleInheritance");
        post(conn);
        ignoreBody(conn);
    }

    private void permissions_addRoleAssignments(final String permissionsUrl, final int roledefid, Collection<User> principals) throws IOException, JSONException {
        for (User principal : principals) {
            URLConnection conn = connect(permissionsUrl + "/RoleAssignments/AddRoleAssignment(principalid=" + principal.getSPUser() + ",roledefid=" + roledefid + ")");
            post(conn);
            ignoreBody(conn);
        }
    }

    private int web_roleDefinitions_getByName_id(String name) throws JSONException, IOException {
        URLConnection conn = connect("Web/RoleDefinitions/GetByName('" + name + "')/Id");
        return getBody(conn).getInt("Id");
    }

    // If there are yearly document sites, they must all have the same structure.
    private String documentLibraryRootFolderUrl;
    private String getStudyFolderRoot(final Study study) throws IOException, JSONException {
        if (documentLibraryRootFolderUrl == null) {
            URLConnection conn = connect("Web/Lists/GetByTitle('" + documentLibraryName + "')/RootFolder/ServerRelativeUrl");
            documentLibraryRootFolderUrl = getBody(conn).getString("ServerRelativeUrl");
        }
        return documentLibraryRootFolderUrl + "/" + study.getAssignedID();
    }

    public String getDocumentSiteUrl(Study study) {
        return documentSiteFullUrlPattern.replace("{year}", study.getYear());
    }

    public void addStudy(final Study study) {
        try {
            openSession(getDocumentSiteUrl(study));
            final List<User> remappedWriterMembers = getRemappedWriterMembers(study), remappedReaderMembers = getRemappedReaderMembers(study);
            final int readRole = web_roleDefinitions_getByName_id("Read"), writeRole = web_roleDefinitions_getByName_id("Contribute");

            // Add top-level study folder.
            final String studyFolder = getStudyFolderRoot(study);
            web_folders_add(studyFolder);
            // Set permissions so the members can browse the top-level study folder.
            final String studyFolderPermissions = getFolderPermissionsUrl(studyFolder);
            permissions_breakRoleInheritance(studyFolderPermissions);
            permissions_addRoleAssignments(studyFolderPermissions, readRole, remappedReaderMembers);
            permissions_addRoleAssignments(studyFolderPermissions, readRole, remappedWriterMembers);

            // Add Stage 1 folder under the study.
            final String stage1Folder = studyFolder + "/" + getStageFolder(study.getStudyStatusCollection().iterator().next());
            web_folders_add(stage1Folder);
            // Set permissions so the authors can write to the draft/submission stage folder.
            final String stage1FolderPermissions = getFolderPermissionsUrl(stage1Folder);
            permissions_breakRoleInheritance(stage1FolderPermissions);
            permissions_addRoleAssignments(stage1FolderPermissions, readRole, remappedReaderMembers);
            permissions_addRoleAssignments(stage1FolderPermissions, writeRole, remappedWriterMembers);

            // Add empty review folder under the study, where review stages will all live.
            final String reviewFolder = studyFolder + "/" + reviewFolderName;
            web_folders_add(reviewFolder);
            permissions_breakRoleInheritance(getFolderPermissionsUrl(reviewFolder));

            // Generated forms and various study documents get one folder per study, too.
            final String teamFolder = studyFolder + "/" + studyTeamFolderName;
            web_folders_add(teamFolder);
            final String teamFolderPermissions = getFolderPermissionsUrl(teamFolder);
            permissions_breakRoleInheritance(teamFolderPermissions);
            permissions_addRoleAssignments(stage1FolderPermissions, readRole, remappedReaderMembers);
            permissions_addRoleAssignments(teamFolderPermissions, writeRole, remappedWriterMembers);
        } catch (IOException | JSONException ex) {
            Logger.getLogger(SharePoint.class.getName()).log(Level.SEVERE, null, ex);
            throw ServiceException.fatal(ex);
        } finally {
            closeSession();
        }
    }

    public void addStageToStudy(final StudyStatus stage) {
        // Study applications live on the base site collection, so we do two different sessions here.
        Collection<String> baseReviewers;
        try {
            openSession(baseSiteFullUrl);
            baseReviewers = getReviewers(stage.getStudy()).reviewers;
        } catch (IOException | JSONException ex) {
            Logger.getLogger(SharePoint.class.getName()).log(Level.SEVERE, null, ex);
            throw ServiceException.fatal(ex);
        } finally {
            closeSession();
        }

        try {
            final Study study = stage.getStudy();
            openSession(getDocumentSiteUrl(study));
    		final Collection<User> remappedWriterMembers = getRemappedWriterMembers(study), remappedReaderMembers = getRemappedReaderMembers(study), remappedReviewers = getRemappedUsers(baseReviewers);
            final int readRole = web_roleDefinitions_getByName_id("Read"), writeRole = web_roleDefinitions_getByName_id("Contribute");
            final String studyFolder = getStudyFolderRoot(study);

            // Reset permissions on the folder for Stage N - 1.
            permissions_resetRoleInheritance(getFolderPermissionsUrl(studyFolder + "/" + getPreviousStageFolder(stage)));
            // Add Stage N folder under the study.
            final String stageNFolder = studyFolder + "/" + getStageFolder(stage);
            web_folders_add(stageNFolder);
            // Set permissions so the authors can write to the draft/submission stage folder.
            final String stageNFolderPermissions = getFolderPermissionsUrl(stageNFolder);
            permissions_breakRoleInheritance(stageNFolderPermissions);
            permissions_addRoleAssignments(stageNFolderPermissions, readRole, remappedReaderMembers);
            permissions_addRoleAssignments(stageNFolderPermissions, writeRole, remappedWriterMembers);
            permissions_addRoleAssignments(stageNFolderPermissions, readRole, remappedReviewers);
        } catch (IOException | JSONException ex) {
            Logger.getLogger(SharePoint.class.getName()).log(Level.SEVERE, null, ex);
            throw ServiceException.fatal(ex);
        } finally {
            closeSession();
        }
    }

    public void updateFolderPermissions(final Study study) {
        // Study applications live on the base site collection, so we do two different sessions here.
        Collection<String> baseReviewers;
        try {
            openSession(baseSiteFullUrl);
            baseReviewers = getReviewers(study).reviewers;
        } catch (IOException | JSONException ex) {
            Logger.getLogger(SharePoint.class.getName()).log(Level.SEVERE, null, ex);
            throw ServiceException.fatal(ex);
        } finally {
            closeSession();
        }

        try {
            openSession(getDocumentSiteUrl(study));
            final List<User> remappedWriterMembers = getRemappedWriterMembers(study), remappedReaderMembers = getRemappedReaderMembers(study), remappedReviewers = getRemappedUsers(baseReviewers);
            final int readRole = web_roleDefinitions_getByName_id("Read"), writeRole = web_roleDefinitions_getByName_id("Contribute");

            // Update permissions of the top-level study folder.
            final String studyFolder = getStudyFolderRoot(study);
            final String studyFolderPermissions = getFolderPermissionsUrl(studyFolder);
            permissions_resetRoleInheritance(studyFolderPermissions); // reset +
            permissions_breakRoleInheritance(studyFolderPermissions); // break = clear
            permissions_addRoleAssignments(studyFolderPermissions, readRole, remappedReaderMembers);
            permissions_addRoleAssignments(studyFolderPermissions, readRole, remappedWriterMembers);
            permissions_addRoleAssignments(studyFolderPermissions, readRole, remappedReviewers);

            // Find latest Stage N folder under the study.
            StudyStatus latestStage = null;
            for (StudyStatus ss : study.getStudyStatusCollection())
                latestStage = ss;
            final String stageNFolder = studyFolder + "/" + getStageFolder(latestStage);
            // Set permissions so the authors can write to the draft/submission stage folder.
            final String stageNFolderPermissions = getFolderPermissionsUrl(stageNFolder);
            permissions_breakRoleInheritance(stageNFolderPermissions);
            permissions_addRoleAssignments(stageNFolderPermissions, readRole, remappedReaderMembers);
            permissions_addRoleAssignments(stageNFolderPermissions, writeRole, remappedWriterMembers);
            permissions_addRoleAssignments(stageNFolderPermissions, readRole, remappedReviewers);

            // Update the reviewers' folder's permissions.
            final String reviewFolder = studyFolder + "/" + reviewFolderName;
            final String reviewFolderPermissions = getFolderPermissionsUrl(reviewFolder);
            permissions_resetRoleInheritance(reviewFolderPermissions); // reset + 
            permissions_breakRoleInheritance(reviewFolderPermissions); // break = clear
            permissions_addRoleAssignments(reviewFolderPermissions, writeRole, remappedReviewers);

            // Update the study team's folder's permissions.
            final String teamFolder = studyFolder + "/" + studyTeamFolderName;
            final String teamFolderPermissions = getFolderPermissionsUrl(teamFolder);
            permissions_resetRoleInheritance(teamFolderPermissions); // reset +
            permissions_breakRoleInheritance(teamFolderPermissions); // break = clear
            permissions_addRoleAssignments(teamFolderPermissions, readRole, remappedReaderMembers);
            permissions_addRoleAssignments(teamFolderPermissions, readRole, remappedReviewers);
            permissions_addRoleAssignments(teamFolderPermissions, writeRole, remappedWriterMembers);
        } catch (IOException | JSONException ex) {
            Logger.getLogger(SharePoint.class.getName()).log(Level.SEVERE, null, ex);
            throw ServiceException.fatal(ex);
        } finally {
            closeSession();
        }
    }

    public URI uploadStudyTeamFile(Study study, File file, String asFileName) {
        try {
            final String siteUrl = getDocumentSiteUrl(study);
            openSession(siteUrl);
            final String studyFolder = getStudyFolderRoot(study);
            final String formsFolder = studyFolder + "/" + studyTeamFolderName;
            web_folders_files_add(formsFolder, file, asFileName);
            return new URI(siteUrl + "/" + (documentLibraryName + "/" + study.getAssignedID() + "/" + studyTeamFolderName + "/" + asFileName).replace(" ", "%20"));
        } catch (IOException | JSONException | URISyntaxException ex) {
            Logger.getLogger(SharePoint.class.getName()).log(Level.SEVERE, null, ex);
            throw ServiceException.fatal(ex);
        } finally {
            closeSession();
        }
    }

    private static String getStageFolder(StudyStatus stage) {
        // Stages owned by the authors go in the study folder.
        // XXX Just creating folders for all stages.
        return String.valueOf(stage.getStageCounter());
    }

    private static String getPreviousStageFolder(StudyStatus stage) {
        return String.valueOf(stage.getStageCounter() - 1);
    }

    private List<User> getRemappedWriterMembers(Study study) throws IOException, JSONException {
        List<User> members = new LinkedList<>();
        StudyStatus initialStage = study.getStudyStatusCollection().iterator().next();
        members.add(web_ensureUser(initialStage.getCreatedBy().getLoginName()));
        for (Study_Members sm : study.getStudyMembersCollection())
            if (sm.isActiveMember() && sm.getUser() != null && sm.isTeamMember() && sm.isWritePermissions())
                members.add(web_ensureUser(sm.getUser().getLoginName()));
        return members;
    }

    private List<User> getRemappedReaderMembers(Study study) throws IOException, JSONException {
        List<User> members = new LinkedList<>();
        for (Study_Members sm : study.getStudyMembersCollection())
            if (sm.isActiveMember() && sm.getUser() != null && sm.isTeamMember() && !sm.isWritePermissions())
                members.add(web_ensureUser(sm.getUser().getLoginName()));
        return members;
    }

    private Integer web_lists_getByTitle_application_item_id(Study study) throws IOException, JSONException {
		URLConnection conn = connect("Web/Lists/GetByTitle('" + applicationListName + "')/Items?"
                + "$filter=Study_ID eq '" + study.getAssignedID() + "'&"
                + "$select=Id");
        JSONArray results = getBody(conn).getJSONArray("results");
        if (results.length() == 0)
            return null;
        else
            return results.getJSONObject(0).getInt("Id");
    }

    private JSONArray web_siteGroups_id_users(int spGroupId) throws IOException, JSONException {
        URLConnection conn = connect("Web/SiteGroups(" + spGroupId + ")/Users");
        return getBody(conn).getJSONArray("results");
    }

    private JSONObject web_lists_getByTitle_application_item(Integer spListItem) throws IOException, JSONException {
        URLConnection conn = connect("Web/Lists/GetByTitle('" + applicationListName + "')/Items(" + spListItem + ")");
        return getBody(conn);
    }

    private JSONObject web_lists_getByTitle_facility_item(int spListItem) throws IOException, JSONException {
        URLConnection conn = connect("Web/Lists/GetByTitle('" + facilityListName + "')/Items(" + spListItem + ")");
        return getBody(conn);
    }

    private JSONObject web_lists_getByTitle_committee_item(int spListItem) throws IOException, JSONException {
        URLConnection conn = connect("Web/Lists/GetByTitle('" + committeeListName + "')/Items(" + spListItem + ")");
        return getBody(conn);
    }

    /**
     * Retrieve the login names for all users in the reviewer columns of the Application list item,
     * or the cached values thereof.  If we notice a change, we update the folder permissions, too.
     */
    public SortedSet<String> getReviewerLoginNames(Study study) throws IOException, JSONException {
        GetReviewersResults res;
        try {
            openSession(baseSiteFullUrl);
            res = getReviewers(study);
        } catch (IOException | JSONException ex) {
            Logger.getLogger(SharePoint.class.getName()).log(Level.SEVERE, null, ex);
            throw ServiceException.fatal(ex);
        } finally {
            closeSession();
        }

        if (res.hasChanged)
            updateFolderPermissions(study);

        return res.reviewers;
    }

    /**
     * Retrieve login name for the assigned IRB administrator (or cached value.), and update folder
     * permissions upon noticing a change.
     */
    public String getAdministratorLoginName(Study study) throws IOException, JSONException {
        GetReviewersResults res;
        try {
            openSession(baseSiteFullUrl);
            res = getReviewers(study);
        } catch (IOException | JSONException ex) {
            Logger.getLogger(SharePoint.class.getName()).log(Level.SEVERE, null, ex);
            throw ServiceException.fatal(ex);
        } finally {
            closeSession();
        }

        if (res.hasChanged)
            updateFolderPermissions(study);

        return res.administrator;
    }


    // Map from Study.studiesID to a hash of all the reviewers' login names so that we know when to propagate permissions.
    private final HashMap<Integer, Integer> studiesIDReviewerSetHashes = new HashMap<>();
    private static class ReviewerCacheEntry {
        private static final long MAX_AGE = 30000; // Let reviewer access control be at most half a minute stale.
        String administrator;
        SortedSet<String> reviewers;
        long time;
        ReviewerCacheEntry(String administrator, SortedSet<String> reviewers, Date now) {
            this.administrator = administrator;
            this.reviewers = Collections.unmodifiableSortedSet(reviewers);
            this.time = now.getTime();
        }

        boolean isExpired(Date now) {
            return now.getTime() - time > MAX_AGE;
        }
    }
    private final WeakHashMap<Study, ReviewerCacheEntry> studyReviewerCache = new WeakHashMap<>();
    private static class GetReviewersResults {
        String administrator;
        SortedSet<String> reviewers;
        boolean hasChanged;
        GetReviewersResults(String administrator, SortedSet<String> reviewers, boolean hasChanged) {
            this.administrator = administrator;
            this.reviewers = reviewers;
            this.hasChanged = hasChanged;
        }
    }
    private GetReviewersResults getReviewers(Study study) throws IOException, JSONException {
        final Date now = new Date();
        ReviewerCacheEntry cacheEntry = studyReviewerCache.get(study);
        if (cacheEntry != null && !cacheEntry.isExpired(now))
            return new GetReviewersResults(cacheEntry.administrator, cacheEntry.reviewers, false);

        if (study.getSPListItem() == null) { // XXX FOR NOW, front-end is writing to the applications list(s) for us
            study.setSPListItem(web_lists_getByTitle_application_item_id(study));
            if (study.getSPListItem() == null)
                return new GetReviewersResults(null, new TreeSet<String>(), false);
            study.setSynchronizedTimestamp(now);
        }

        // Add reviewers immediately present in the Application.
        // The Assigned IRB Administrator is listed among the reviewer columns, as well.
        final JSONObject application = web_lists_getByTitle_application_item(study.getSPListItem());
        String administrator = null;
        final SortedSet<String> reviewers = new TreeSet<>();
        for (final String reviewerColumn : applicationListReviewerColumnNames) {
            if (!application.isNull(reviewerColumn)) {
                final Object column = application.get(reviewerColumn);
                if (column instanceof JSONObject) { // Multi- or single-value of SP.User?
                    final JSONArray ids = ((JSONObject) column).getJSONArray("results");
                    for (int i = 0; i < ids.length(); i++)
                        reviewers.add(web_getUserById_loginName(ids.getInt(i)));
                } else {
                    String login = web_getUserById_loginName(((java.lang.Number) column).intValue());
                    reviewers.add(login);
                    if (reviewerColumn.equals(applicationListAdministratorColumnName))
                        administrator = login;
                }
            }
        }
        // Descend into the Application's Facility and add the Facility Liaison as a reviewer.
        if (!application.isNull(applicationListFacilityLookupColumnName)) {
            final JSONObject facility = web_lists_getByTitle_facility_item(application.getInt(applicationListFacilityLookupColumnName));
            if (!facility.isNull(facilityListLiaisonColumnName))
                reviewers.add(web_getUserById_loginName(facility.getInt(facilityListLiaisonColumnName)));
            // Finally, visit the Facility's Committee and for each group column, add all the users in that group.
            final JSONObject committee = web_lists_getByTitle_committee_item(facility.getInt(facilityListCommitteeLookupColumnName));
            for (final String groupColumn : committeeListGroupColumnNames) {
                final JSONArray users = web_siteGroups_id_users(committee.getInt(groupColumn));
                for (int i = 0; i < users.length(); i++) {
                    String name = users.getJSONObject(i).getString("LoginName");
                    if (name.startsWith(DOMAIN_USER_PREFIX)) // should always be present
                        name = name.substring(DOMAIN_USER_PREFIX.length());
                    reviewers.add(name);
                }
            }
        }

        // Update the reviewer cache and then see if the reviewers may have changed from before by comparing the hash.
        studyReviewerCache.put(study, cacheEntry = new ReviewerCacheEntry(administrator, reviewers, now));
        Integer newHash = cacheEntry.reviewers.hashCode();
        Integer oldHash = studiesIDReviewerSetHashes.put(study.getStudiesID(), newHash);
        return new GetReviewersResults(administrator, cacheEntry.reviewers, !newHash.equals(oldHash));
    }

    private List<User> getRemappedUsers(Collection<String> loginNames) throws IOException, JSONException {
        List<User> remapped = new ArrayList<>(loginNames.size());
        for (String user : loginNames)
            remapped.add(web_ensureUser(user));
        return remapped;
    }
}
